"""Neuroimaging statistics toolbox."""
__version__ = "0.1.3"

import hippomaps.config
import hippomaps.utils
import hippomaps.plotting
import hippomaps.stats
import hippomaps.fetcher

__all__ = ['config', 'utils', 'plotting', 'stats', 'fetcher']

